from . import utils
from .base import Expression
from .constants import POS, RESULT, STATUS


class Opt(Expression):
    num_blocks = 1

    def __init__(self, expr):
        self.expr = expr

    def __str__(self):
        return f'Opt({self.expr})'

    def always_succeeds(self):
        return True

    def can_partially_succeed(self):
        return False

    def _compile(self, out):
        backtrack = out.var('backtrack', POS)
        with utils.if_fails(out, self.expr):
            out += POS << backtrack
            out += RESULT << None
            out += STATUS << True



from .base import Expression
from .constants import RESULT, STATUS
from .str import Str


class Fail(Expression):
    num_blocks = 0

    def __init__(self, message=None):
        if isinstance(message, Str):
            message = message.value
        self.message = message

    def __str__(self):
        return 'Fail()' if self.message is None else f'Fail({self.message!r})'

    def _compile(self, out):
        out += RESULT << self.error_func()
        out += STATUS << False

    def complain(self):
        return 'Failed' if self.message is None else str(self.message)



from outsourcer import Code

from . import utils
from .base import Expression
from .constants import POS, RESULT, STATUS, TEXT
from .seq import Seq


class Class(Expression):
    has_params = True

    is_tagged = False
    is_commented = False

    num_blocks = 2

    def __init__(self, name, params, fields, is_ignored=False):
        self.name = name
        self.params = params
        self.fields = fields
        self.is_ignored = is_ignored
        self.extra_id = None

    def __str__(self):
        params = '' if self.params is None else f'({", ".join(self.params)})'
        fields = ''.join(f'    {x.name}: {x.expr}\n' for x in self.fields)
        return f'class {self.name}{params} {{\n{fields}}}'

    def always_succeeds(self):
        return all(x.expr.always_succeeds() for x in self.fields)

    def _compile(self, out):
        parse_func = Code(f'{utils.implementation_name(self.name)}')
        field_names = [x.name for x in self.fields]

        with out.global_section():
            with out.CLASS(self.name, 'Node'):
                self._compile_class_body(out, parse_func, field_names)

            with out.DEF(parse_func, [str(TEXT), str(POS)] + (self.params or [])):
                exprs = (x.expr for x in self.fields)
                seq = Seq(*exprs, names=field_names, constructor=self.name)
                seq.program_id = self.extra_id
                seq.compile(out)
                out.YIELD((STATUS, RESULT, POS))

    def _compile_class_body(self, out, parse_func, field_names):
        out.add_docstring(str(self))
        out += Code('_fields') << tuple(field_names)
        out.add_newline()

        with out.DEF('__init__', ['self'] + field_names):
            out += Code('Node.__init__(self)')
            for name in field_names:
                out += Code(f'self.{name} = {name}')

        with out.DEF('__repr__', ['self']):
            values = ', '.join(f'{x}={{self.{x}!r}}' for x in field_names)
            out.RETURN(Code(f"f'{self.name}({values})'"))

        out += Code('@staticmethod')
        if self.params:
            with out.DEF('parse', self.params):
                _closure, _ParseFunction = Code('_closure'), Code('_ParseFunction')
                args = tuple(Code(x) for x in self.params)
                out += _closure << _ParseFunction(parse_func, args, {})
                out.RETURN(Code(
                    'lambda text, pos=0, fullparse=True:'
                    ' _run(text, pos, _closure, fullparse)'
                ))
        else:
            with out.DEF('parse', ['text', 'pos=0', 'fullparse=True']):
                out.RETURN(Code(f'_run(text, pos, {parse_func}, fullparse)'))



from outsourcer import Code

from . import utils
from .base import Expression
from .constants import POS, RESULT, STATUS, TEXT


class Byte(Expression):
    def __init__(self, value):
        if not isinstance(value, int):
            raise TypeError(f'Expected int. Received: {type(value)}.')

        if not (0x00 <= value <= 0xFF):
            raise ValueError(
                f'Expected integer in the range [0, 255]. Received: {value!r}.'
            )

        self.value = value
        self.skip_ignored = False
        self.num_blocks = 1

    def __str__(self):
        return hex(self.value)

    def always_succeeds(self):
        return False

    def can_partially_succeed(self):
        return False

    def argumentize(self, out):
        wrap = Code('_wrap_byte_literal')
        value = Expression.argumentize(self, out)
        return out.var('arg', wrap(self.value, value))

    def _compile(self, out):
        LEN = Code('len')
        has_byte = POS < LEN(TEXT)
        is_match = TEXT[POS] == self.value

        with out.IF(Code(has_byte, ' and ', is_match)):
            out += RESULT << self.value
            end = POS + 1
            out += POS << (utils.skip_ignored(end) if self.skip_ignored else end)
            out += STATUS << True

        with out.ELSE():
            out += RESULT << self.error_func()
            out += STATUS << False

    def complain(self):
        return f'Expected to match the byte value {hex(self.value)}'



from outsourcer import Code

from . import utils
from .base import Expression
from .constants import RESULT


class Let(Expression):
    defines_local = True
    num_blocks = 1

    def __init__(self, name, expr, body):
        self.name = name
        self.expr = expr
        self.body = body

    def __str__(self):
        return f'let {self.name} = {self.expr} in\n{self.body}'

    def operand_string(self):
        return f'({self})'

    def always_succeeds(self):
        return (self.expr.always_succeeds()
            and self.body.always_succeeds())

    def _compile(self, out):
        with utils.if_succeeds(out, self.expr):
            out += Code(self.name) << RESULT
            self.body.compile(out)



from . import utils
from .base import Expression
from .constants import BREAK, RESULT


class Discard(Expression):
    num_blocks = 2

    def __init__(self, expr1, expr2, discard_left=True):
        self.expr1 = expr1
        self.expr2 = expr2
        self.discard_left = discard_left

    def __str__(self):
        op = '>>' if self.discard_left else '<<'
        return utils.infix_str(self.expr1, op, self.expr2)

    def operand_string(self):
        return f'({self})'

    def always_succeeds(self):
        return (self.expr1.always_succeeds()
            and self.expr2.always_succeeds())

    def _compile(self, out):
        with utils.breakable(out):
            with utils.if_fails(out, self.expr1):
                out += BREAK

            if self.discard_left:
                self.expr2.compile(out)
            else:
                staging = out.var('staging', RESULT)
                with utils.if_succeeds(out, self.expr2):
                    out += RESULT << staging



from . import utils
from .base import Expression
from .constants import RESULT


class Apply(Expression):
    num_blocks = 2

    def __init__(self, expr1, expr2, apply_left=False):
        self.expr1 = expr1
        self.expr2 = expr2
        self.apply_left = apply_left

    def __str__(self):
        op = '<|' if self.apply_left else '|>'
        return utils.infix_str(self.expr1, op, self.expr2)

    def operand_string(self):
        return f'({self})'

    def always_succeeds(self):
        return self.expr1.always_succeeds() and self.expr2.always_succeeds()

    def _compile(self, out):
        with utils.if_succeeds(out, self.expr1):
            first = out.var('func' if self.apply_left else 'arg', RESULT)
            with utils.if_succeeds(out, self.expr2):
                result = first(RESULT) if self.apply_left else RESULT(first)
                out += RESULT << result



from string import Template

from outsourcer import CodeBuilder, Code, Val

from . import expressions as ex
from .expressions import (
    TEXT, POS, Choice, Class, Ref, Right, Rule, Skip, visit
)


def generate_source_code(docstring, nodes):
    out = CodeBuilder()
    out.add_docstring(docstring)
    out += Code(_program_setup)

    # Collect all the rules and stuff.
    rules, ignored = [], []
    start_rule = None

    for node in nodes:
        # Just add Python sections directly to the program.
        if isinstance(node, (ex.PythonExpression, ex.PythonSection)):
            out += Code(node.source_code)
            continue

        rules.append(node)

        if node.is_ignored:
            ignored.append(node)

        if start_rule is None and node.name and node.name.lower() == 'start':
            start_rule = node

    if start_rule is not None and start_rule.is_ignored:
        raise Exception(
            f'The {start_rule!r} rule may not have the "ignored" modifier.'
        )

    if not rules:
        raise Exception('Expected one or more grammar rules.')

    visited_names = set()
    for rule in rules:
        if rule.name is not None and rule.name.startswith('_'):
            raise Exception(
                'Grammar rule names must start with a letter. Found a rule that'
                f' starts with an underscore: "{rule.name}". '
            )

        if not rule.name:
            rule.name = f'_anonymous_{id(rule)}'

        if rule.name in visited_names:
            raise Exception(
                'Each grammar rule must have a unique name. Found two or more'
                f' rules named "{rule.name}".'
            )
        visited_names.add(rule.name)

    if ignored:
        # Create a rule called "_ignored" that skips all the ignored rules.
        refs = [Ref(x.name) for x in ignored]
        rules.append(Rule('_ignored', None, Skip(*refs), 'ignored'))

        # If we have a start rule, then update its expression to skip ahead past
        # any leading ignored stuff.
        if isinstance(start_rule, Class):
            first_rule = start_rule.fields[0] if start_rule.fields else None
        else:
            first_rule = start_rule

        if first_rule:
            assert isinstance(first_rule, Rule)
            impl_name = ex.implementation_name('_ignored')
            first_rule.expr = Right(Ref(impl_name), first_rule.expr)

        # Update the "skip_ignored" flag of each literal.
        def _set_skip_ignored(expr):
            if hasattr(expr, 'skip_ignored'):
                expr.skip_ignored = True

        for rule in rules:
            if not rule.is_ignored:
                visit(rules, _set_skip_ignored)

    _assign_ids(rules)
    _update_local_references(rules)
    _update_rule_references(rules)

    default_rule = start_rule or rules[0]

    out += Code(Template(_main_template).substitute(
        CALL=ex.CALL,
        start=ex.implementation_name(default_rule.name),
    ))

    error_delegates = {}
    def set_error_delegate(expr):
        if not isinstance(expr, Choice):
            return
        real, fail = [], []
        for option in expr.exprs:
            if isinstance(option, ex.Fail):
                fail.append(option)
            else:
                real.append(option)
        if not real or not fail:
            return
        delegate = Choice(*real)
        error_delegates[fail[-1].program_id] = Choice(*real)

    for rule in rules:
        visit(rule, set_error_delegate)

    visited = set()
    def maybe_compile_error_message(out, rule, expr):
        if not hasattr(expr, 'complain') or expr.program_id in visited:
            return

        visited.add(expr.program_id)
        if expr.always_succeeds():
            return

        with out.global_section():
            TITLE, LINE, COL = Code('title'), Code('line'), Code('col')

            with out.DEF(str(expr.error_func()), [str(TEXT), str(POS)]):
                with out.IF(Code('len')(TEXT) <= POS):
                    out += TITLE << 'Unexpected end of input.'
                    out += LINE << None
                    out += COL << None

                with out.ELSE():
                    out += (LINE, COL) << Code('_get_line_and_column')(TEXT, POS)
                    out += Code('excerpt') << Code('_extract_excerpt')(TEXT, POS, COL)
                    out += TITLE << Code(
                        r"f'Error on line {line}, column {col}:\n{excerpt}\n'"
                    )

                delegate = error_delegates.get(expr.program_id, expr)
                out.extend([
                    Code('details = ('),
                    Val(f'Failed to parse the {rule.name!r} rule, at the expression:\n'),
                    Val(f'    {str(delegate)}\n\n'),
                    Val(expr.complain()),
                    Code(')'),
                    Code('raise ParseError', (TITLE + Code('details'), POS, LINE, COL)),
                ])

    for rule in rules:
        visit(rule, lambda x: x.precompile(out))

    out.add_newline()

    for rule in rules:
        rule.compile(out)
        visit(rule, lambda x: maybe_compile_error_message(out, rule, x))

    return out


def _assign_ids(rules):
    next_id = 1

    def assign_id(node):
        nonlocal next_id
        node.program_id = next_id
        next_id += 1
        if isinstance(node, Class):
            node.extra_id = next_id
            next_id += 1

    visit(rules, assign_id)


def _update_local_references(rules):
    counter = ex.SymbolCounter()

    def previsit(node):
        counter.previsit(node)
        if node.is_reference and counter.is_bound(node.name):
            node.is_local = True

    visit(rules, previsit, counter.postvisit)


def _update_rule_references(rules):
    rule_names = set()
    for rule in rules:
        if isinstance(rule, (Class, Rule)):
            rule_names.add(rule.name)

    def check_refs(node):
        if isinstance(node, Ref) and node.name in rule_names and not node.is_local:
            node._resolved = ex.implementation_name(node.name)

    visit(rules, check_refs)


_program_setup = r'''
from collections import namedtuple as _nt
from re import compile as _compile_re, IGNORECASE as _IGNORECASE

class Node:
    _fields = ()

    def __init__(self):
        self._metadata = _Metadata()

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for field in self._fields:
            if getattr(self, field) != getattr(other, field):
                return False
        return True

    def _asdict(self):
        return {k: getattr(self, k) for k in self._fields}

    def _replace(self, **kw):
        for field in self._fields:
            if field not in kw:
                kw[field] = getattr(self, field)
        result = self.__class__(**kw)
        result._metadata.update(self._metadata)
        return result


class _Metadata:
    def __init__(self, **fields):
        object.__setattr__(self, '_fields', fields)

    def __getattr__(self, name):
        return self._fields.get(name)

    def __setattr__(self, name, value):
        self._fields[name] = value

    def __len__(self):
        return len(self._fields)

    def copy(self):
        return _Metadata(**self._fields)

    def update(self, other):
        self._fields.update(other._fields)


class Rule:
    def __init__(self, name, parse, definition):
        self.name = name
        self.parse = parse
        self.definition = definition

    def __repr__(self):
        return (f'Rule(name={self.name!r}, parse={self.parse.__name__},'
            f' definition={self.definition!r})')
'''


_main_template = r'''
class SourcerError(Exception):
    """Common superclass for ParseError and PartialParseError."""


class ParseError(SourcerError):
    def __init__(self, message, index, line, column):
        super().__init__(message)
        self.position = _Position(index, line, column)


class PartialParseError(SourcerError):
    def __init__(self, partial_result, last_position, excerpt):
        super().__init__('Incomplete parse. Unexpected input on line'
            f' {last_position.line}, column {last_position.column}:\n{excerpt}')
        self.partial_result = partial_result
        self.last_position = last_position


class Infix(Node):
    _fields = ('left', 'operator', 'right')

    def __init__(self, left, operator, right):
        Node.__init__(self)
        self.left = left
        self.operator = operator
        self.right = right

    def __repr__(self):
        return f'Infix({self.left!r}, {self.operator!r}, {self.right!r})'


class Postfix(Node):
    _fields = ('left', 'operator')

    def __init__(self, left, operator):
        Node.__init__(self)
        self.left = left
        self.operator = operator

    def __repr__(self):
        return f'Postfix({self.left!r}, {self.operator!r})'


class Prefix(Node):
    _fields = ('operator', 'right')

    def __init__(self, operator, right):
        Node.__init__(self)
        self.operator = operator
        self.right = right

    def __repr__(self):
        return f'Prefix({self.operator!r}, {self.right!r})'


def parse(text, pos=0, fullparse=True):
    return _run(text, pos, $start, fullparse)


_PositionInfo = _nt('_PositionInfo', 'start, end')

_Position = _nt('_Position', 'index, line, column')


class _ParseFunction(_nt('_ParseFunction', 'func, args, kwargs')):
    def __call__(self, _text, _pos):
        return self.func(_text, _pos, *self.args, **dict(self.kwargs))


class _StringLiteral(str):
    def __call__(self, _text, _pos):
        return self._parse_function(_text, _pos)


def _wrap_string_literal(string_value, parse_function):
    result = _StringLiteral(string_value)
    result._parse_function = parse_function
    return result


class _ByteLiteral(int):
    def __call__(self, _text, _pos):
        return self._parse_function(_text, _pos)


def _wrap_byte_literal(byte_value, parse_function):
    result = _ByteLiteral(byte_value)
    result._parse_function = parse_function
    return result


def _run(text, pos, start, fullparse):
    memo = {}
    result = None

    key = ($CALL, start, pos)
    gtor = start(text, pos)
    stack = [(key, gtor)]

    while stack:
        key, gtor = stack[-1]
        result = gtor.send(result)

        if result[0] != $CALL:
            stack.pop()
            memo[key] = result
        elif result in memo:
            result = memo[result]
        else:
            gtor = result[1](text, result[2])
            stack.append((result, gtor))
            result = None

    if result[0]:
        return _finalize_parse_info(text, result[1], result[2], fullparse)
    else:
        pos = result[2]
        message = result[1](text, pos)
        raise ParseError(message, pos)


def visit(node):
    visited = set()
    stack = [node]
    while stack:
        node = stack.pop()

        if isinstance(node, (list, tuple)):
            stack.extend(node)

        elif isinstance(node, dict):
            stack.extend(node.values())

        elif isinstance(node, Node):
            node_id = id(node)
            if node_id in visited:
                continue
            visited.add(node_id)

            yield node

            if hasattr(node, '_fields'):
                stack.extend(getattr(node, x) for x in node._fields)


_Traversing = _nt('_Traversing', 'parent, field, child, is_finished')


def traverse(node):
    visited = set()
    stack = [_Traversing(parent=None, field=None, child=node, is_finished=False)]
    while stack:
        traversing = stack.pop()

        if traversing.is_finished:
            yield traversing
            continue

        child = traversing.child
        child_id = id(child)

        if child_id in visited:
            continue

        visited.add(child_id)
        stack.append(traversing._replace(is_finished=True))
        yield traversing

        def extend(items):
            stack.extend(reversed(list(items)))

        if isinstance(child, (list, tuple)):
            extend(
                _Traversing(parent=child, field=i, child=x, is_finished=False)
                for i, x in enumerate(child)
            )

        elif isinstance(child, dict):
            extend(
                _Traversing(parent=child, field=k, child=v, is_finished=False)
                for k, v in child.items()
            )

        elif isinstance(child, Node) and hasattr(child, '_fields'):
            extend(
                _Traversing(
                    parent=child,
                    field=x,
                    child=getattr(child, x),
                    is_finished=False,
                )
                for x in child._fields
            )


def transform(node, *callbacks):
    if not callbacks:
        return node

    if len(callbacks) == 1:
        callback = callbacks[0]
    else:
        def callback(node):
            for f in callbacks:
                node = f(node)
            return node

    return _transform(node, callback)


def _transform(node, callback):
    if isinstance(node, list):
        return [_transform(x, callback) for x in node]

    if not isinstance(node, Node):
        return node

    updates = {}
    for field in node._fields:
        was = getattr(node, field)
        now = _transform(was, callback)
        if now is not was:
            updates[field] = now
            if isinstance(was, Node) and isinstance(now, Node) and not now._metadata:
                now._metadata.update(was._metadata)

    if updates:
        node = node._replace(**updates)

    return callback(node)


def _finalize_parse_info(text, nodes, pos, fullparse):
    line_numbers, column_numbers = _map_index_to_line_and_column(text)

    for node in visit(nodes):
        pos_info = node._metadata.position_info
        if pos_info:
            start, end = pos_info
            end -= 1
            node._metadata.position_info = _PositionInfo(
                start=_Position(start, line_numbers[start], column_numbers[start]),
                end=_Position(end, line_numbers[end], column_numbers[end]),
            )

    if fullparse and pos < len(text):
        line, col = line_numbers[pos], column_numbers[pos]
        position = _Position(pos, line, col)
        excerpt = _extract_excerpt(text, pos, col)
        raise PartialParseError(nodes, position, excerpt)

    return nodes


def _extract_excerpt(text, pos, col):
    if isinstance(text, bytes):
        return repr(text[max(0, pos - 1) : pos + 2])

    start = pos - (col - 1)
    match = _compile_re('\n').search(text, pos + 1)
    end = len(text) if match is None else match.start()

    if end - start < 96:
        return text[start : end] + _caret_at(col - 1)

    if col < 60:
        # Chop the line off at the end.
        return text[start : start + 90] + ' ...' + _caret_at(col - 1)

    elif end - pos < 40:
        # Chop the line off at the start.
        return '... ' + text[end - 90 : end] + _caret_at(pos - (end - 90) + 4)

    else:
        # Chop the line off at both ends.
        return '... ' + text[pos - 42 : pos + 42] + ' ...' + _caret_at(42 + 4)


def _caret_at(index):
    return '\n' + (' ' * index) + '^'


def _get_line_and_column(text, pos):
    line_numbers, column_numbers = _map_index_to_line_and_column(text)
    return line_numbers[pos], column_numbers[pos]


def _map_index_to_line_and_column(text):
    line_numbers = []
    column_numbers = []

    current_line = 1
    current_column = 0

    for c in text:
        if c == '\n':
            current_line += 1
            current_column = 0
        else:
            current_column += 1
        line_numbers.append(current_line)
        column_numbers.append(current_column)

    return line_numbers, column_numbers
'''



from collections import defaultdict
from outsourcer import Code
from .constants import POS, RESULT, STATUS, TEXT


class Expression:
    defines_local = False
    has_params = False

    is_commented = True
    is_reference = False
    is_tagged = True

    def always_succeeds(self):
        return False

    def can_partially_succeed(self):
        # By default, assume that if you don't always succeed, then you can
        # partially succeed.
        return not self.always_succeeds()

    def precompile(self, out):
        pass

    def compile(self, out):
        if not out.has_available_blocks(self.num_blocks):
            func, params = self.functionalize(out, is_generator=False)
            out += (STATUS, RESULT, POS) << func(*params)
            return

        if self.is_tagged:
            out.add_comment(f'Begin {self.__class__.__name__}')

        if self.is_commented:
            out.add_comment(str(self))

        self._compile(out)

        if self.is_tagged:
            out.add_comment(f'End {self.__class__.__name__}')

    def error_func(self):
        return Code(f'_raise_error{self.program_id}')

    def operand_string(self):
        return str(self)

    def argumentize(self, out):
        func, params = self.functionalize(out, is_generator=True)
        if len(params) <= 2:
            return func
        else:
            _ParseFunction = Code('_ParseFunction')
            value = _ParseFunction(func, tuple(params[2:]), ())
            return out.var('arg', value)

    def functionalize(self, out, is_generator=False):
        name = f'_parse_function_{self.program_id}'
        params = [str(TEXT), str(POS)] + list(sorted(self.freevars()))

        with out.global_section():
            with out.DEF(name, params):
                self.compile(out)
                method = out.YIELD if is_generator else out.RETURN
                method((STATUS, RESULT, POS))

        return Code(name), [Code(x) for x in params]

    def freevars(self):
        counter = SymbolCounter()
        visit(self, counter.previsit, counter.postvisit)
        return counter.freevars


def visit(expr, previsitor, postvisitor=None):
    if isinstance(expr, Expression):
        previsitor(expr)

        for child in expr.__dict__.values():
            visit(child, previsitor, postvisitor)

        if postvisitor:
            postvisitor(expr)

    elif isinstance(expr, (list, tuple)):
        for child in expr:
            visit(child, previsitor, postvisitor)


class SymbolCounter:
    def __init__(self):
        self.freevars = set()
        self._counts = defaultdict(int)

    def previsit(self, node):
        if node.defines_local:
            self._counts[node.name] += 1

        if node.has_params and node.params:
            for param in node.params:
                self._counts[param] += 1

        if node.is_reference and node.is_local and not self.is_bound(node.name):
            self.freevars.add(node.name)

    def postvisit(self, node):
        if node.defines_local:
            self._counts[node.name] -= 1

        if node.has_params and node.params:
            for param in node.params:
                self._counts[param] -= 1

    def is_bound(self, name):
        return self._counts[name] > 0



from outsourcer import Code, Yield

from .base import Expression
from .constants import CALL, POS, RESULT, STATUS


class Call(Expression):
    num_blocks = 0

    def __init__(self, func, args):
        self.func = func
        self.args = args

    def __str__(self):
        args = ', '.join(str(x) for x in self.args)
        return f'{self.func}({args})'

    def _compile(self, out):
        args, kwargs = [], []

        for arg in self.args:
            is_kw = isinstance(arg, KeywordArg)
            expr = arg.expr if is_kw else arg
            value = expr.argumentize(out)

            if is_kw:
                kwargs.append((arg.name, value))
            else:
                args.append(value)

        _ParseFunction = Code('_ParseFunction')
        func = _ParseFunction(Code(self.func.resolved), tuple(args), tuple(kwargs))
        func = out.var('func', func)
        out += (STATUS, RESULT, POS) << Yield((CALL, func, POS))


class KeywordArg:
    def __init__(self, name, expr):
        self.name = name
        self.expr = expr

    def __str__(self):
        return f'{self.name}={self.expr}'



from . import utils
from .base import Expression
from .constants import POS, RESULT, STATUS


class Where(Expression):
    num_blocks = 2

    def __init__(self, expr, predicate):
        self.expr = expr
        self.predicate = predicate

    def __str__(self):
        return utils.infix_str(self.expr, 'where', self.predicate)

    def operand_string(self):
        return f'({self})'

    def _compile(self, out):
        with utils.if_succeeds(out, self.expr):
            arg = out.var('arg', RESULT)

            with utils.if_succeeds(out, self.predicate):
                with out.IF(RESULT(arg)):
                    out += RESULT << arg

                with out.ELSE():
                    out += RESULT << self.error_func()
                    out += STATUS << False

    def complain(self):
        return f'Expected to satisfy the predicate: {self.predicate}'



from .base import Expression
from .choice import Choice


class OperatorPrecedence(Expression):
    def __init__(self, atom, *rules):
        self.atom = atom
        self.rules = rules
        self.num_blocks = (rules[-1] if rules else atom).num_blocks

    def __str__(self):
        rules = [self.atom] + list(self.rules)
        lines = ',\n'.join(f'    {x}' for x in rules)
        return f'OperatorPrecedence(\n{lines}\n)'

    def _compile(self, pb):
        prev = self.atom
        for rule in self.rules:
            rule.operand = prev
            prev = rule
        prev.compile(pb)


class OperatorPrecedenceRule(Expression):
    def __init__(self, *operators):
        self.operators = operators[0] if len(operators) == 1 else Choice(*operators)
        self.operand = None

    def __str__(self):
        return f'{self.__class__.__name__}({self.operators})'



import typing

from outsourcer import Code

from . import utils
from .base import Expression
from .constants import POS, RESULT, STATUS, TEXT


class Regex(Expression):
    num_blocks = 1

    def __init__(self, pattern, ignore_case=False):
        if isinstance(pattern, typing.Pattern):
            pattern = pattern.pattern
        if not isinstance(pattern, (bytes, str)):
            raise TypeError('Expected bytes or str')
        self.pattern = pattern
        self.skip_ignored = False
        self.ignore_case = ignore_case

    def __str__(self):
        pattern = self.pattern
        if isinstance(pattern, bytes):
            pattern = pattern.decode('ascii')

        pattern = pattern.replace('\\', '\\\\')
        flag = 'i' if self.ignore_case else ''
        return f'/{pattern}/{flag}'

    def can_partially_succeed(self):
        return False

    def _match_func(self):
        flags = '_IGNORECASE' if self.ignore_case else '0'
        return f'_compile_re({self.pattern!r}, flags={flags}).match'

    def precompile(self, out):
        func = self._match_func()
        if func not in out.state:
            with out.global_section():
                out.state[func] = out.var('matcher', Code(func))

    def _compile(self, out):
        func = out.state[self._match_func()]
        match = out.var('match', func(TEXT, POS))
        end = match.end()

        with out.IF(match):
            out += RESULT << match.group(0)
            out += POS << (utils.skip_ignored(end) if self.skip_ignored else end)
            out += STATUS << True

        with out.ELSE():
            out += RESULT << self.error_func()
            out += STATUS << False

    def complain(self):
        return f'Expected to match the regular expression /{self.pattern}/'



from .discard import Discard
from .list import List


def Left(expr1, expr2):
    return Discard(expr1, expr2, discard_left=False)


def Right(expr1, expr2):
    return Discard(expr1, expr2, discard_left=True)


def Some(expr):
    return List(expr, min_len=1)



import os
import re
import setuptools


def long_description():
    try:
        with open('README.md') as f:
            return f.read()
    except:
        return ''


def install_requires():
    with open('requirements.txt') as f:
        return f.read().splitlines()


def version():
    with open(os.path.join('sourcer', '__init__.py')) as f:
        regex = re.compile('\\n__version__\\s*=\\s*[\'"]+([\\d\\.]+)[\'"]\n')
        return regex.search(f.read()).group(1)


setuptools.setup(
    name='sourcer',
    version=version(),
    author='jvs',
    author_email='vonseg@protonmail.com',
    url='https://github.com/jvs/sourcer',
    description='simple parsing library',
    long_description=long_description(),
    long_description_content_type='text/markdown',
    data_files=[('', ['README.md', 'requirements.txt', 'requirements-dev.txt'])],
    python_requires='>=3.6',
    install_requires=install_requires(),
    packages=['sourcer', 'sourcer.expressions'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Topic :: Software Development :: Interpreters',
        'Topic :: Software Development :: Compilers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
    ],
    platforms='any',
    license='MIT License',
    keywords=['packrat', 'parser', 'peg'],
)



from outsourcer import Code

from . import utils
from .constants import BREAK, POS, RESULT, STATUS
from .precedence import OperatorPrecedenceRule


class LeftAssoc(OperatorPrecedenceRule):
    associativity = 'left'
    num_blocks = 2

    def _compile(self, out):
        is_first = out.var('is_first', True)
        staging = out.var('staging', None)
        operator = out.var('operator')

        with out.WHILE(True):
            with utils.if_fails(out, self.operand):
                out += BREAK

            checkpoint = out.var('checkpoint', POS)

            with out.IF(is_first):
                out += is_first << False
                out += staging << RESULT

            with out.ELSE():
                out += staging << Code('Infix')(staging, operator, RESULT)
                if self.associativity is None:
                    out += BREAK

            with utils.if_fails(out, self.operators):
                out += BREAK

            out += operator << RESULT

        with out.IF_NOT(is_first):
            out += STATUS << True
            out += RESULT << staging
            out += POS << checkpoint


class NonAssoc(LeftAssoc):
    associativity = None


class RightAssoc(OperatorPrecedenceRule):
    associativity = 'right'
    num_blocks = 4

    def _compile(self, out):
        backup = out.var('backup', None)
        prev = out.var('prev', None)

        staging = out.var('staging')
        checkpoint = out.var('checkpoint')

        with out.WHILE(True):
            with utils.if_fails(out, self.operand):
                with out.IF(prev):
                    with out.IF(backup):
                        out += backup.right << prev.left
                        out += RESULT << staging

                    with out.ELSE():
                        out += RESULT << prev.left

                    out += POS << checkpoint
                    out += STATUS << True
                out += BREAK

            out += checkpoint << POS
            operand = out.var('operand', RESULT)

            with utils.if_fails(out, self.operators):
                with out.IF(prev):
                    out += prev.right << operand
                    out += RESULT << staging

                with out.ELSE():
                    out += RESULT << operand

                out += POS << checkpoint
                out += STATUS << True
                out += BREAK

            step = Code('Infix')(operand, RESULT, None)

            with out.IF(prev):
                out += backup << prev
                out += backup.right << prev << step

            with out.ELSE():
                out += staging << prev << step
